import asyncio
import concurrent.futures
import functools
import glob
import json
import logging
import os
import shutil
from datetime import datetime
from pathlib import Path
from typing import Dict, List, Union

from .exceptions import LayerNotFoundInModel
from .utils import download

METADATA_FILENAME = ".metadata.json"

FILE_EXT = [
    "gpml",
    "gpmlz",
    "gpml.gz",
    "dat",
    "pla",
    "shp",
    "geojson",
    "json",
    ".gpkg",
    "gmt",
    "vgp",
]

logger = logging.getLogger("pmm")


class PlateModel:
    """Download and manage files required for a plate reconstruction model.

    👀👇 **LOOK HERE!!!** 👀👇

    Normally you should always use :py:meth:`PlateModelManager.get_model()` to get a :class:`PlateModel` object.
    Create a :class:`PlateModel` object directly only when you don't have Internet connection and would like
    to use the local model files in ``readonly`` mode.
    Do not create a :class:`PlateModel` object directly if you have no idea what's going on.

    .. seealso::

        `Use PlateModel class in readonly mode. <examples.html#use-without-internet>`__
    """

    def __init__(
        self,
        model_name: str,
        model_cfg=None,
        data_dir: str = ".",
        readonly=False,
        timeout=(None, None),
    ):
        """Constructor. Create a :class:`PlateModel` instance.

        :param model_name: The model name of interest.
        :type model_name: str
        :param model_cfg: The model configuration in JSON format.
                          The configuration is either downloaded from the server or
                          loaded from a local file ``.metadata.json``. If you are confused by this parameter,
                          use :py:meth:`PlateModelManager.get_model()` to get a :class:`PlateModel` object instead.
        :param data_dir: The folder path to save the model data.
        :type data_dir: str, default="."
        :param readonly: If this flag is set to ``True``, The :class:`PlateModel` object will use
                         the files in the local folder and will not attempt to
                         download/update the files from the server.
        :type readonly: bool, default=False
        :param timeout: Network connection `timeout parameter <https://requests.readthedocs.io/en/latest/user/advanced/#timeouts>`__.
        """
        self.model_name = model_name.lower()
        self.meta_filename = METADATA_FILENAME
        self._model = model_cfg
        self.readonly = readonly
        self.timeout = timeout

        self.data_dir = data_dir

        self.model_dir = f"{self.data_dir}/{self.model_name}"

        if readonly:
            if not PlateModel.is_model_dir(self.model_dir):
                raise Exception(
                    f"{self.model_dir} must be valid model dir in readonly mode."
                )
            else:
                with open(f"{self.model_dir}/{self.meta_filename}", "r") as f:
                    self._model = json.load(f)

        if not readonly:
            # async and concurrent things
            self.executor = concurrent.futures.ThreadPoolExecutor(max_workers=15)
            self.loop = asyncio.new_event_loop()
            self.run = functools.partial(self.loop.run_in_executor, self.executor)
            asyncio.set_event_loop(self.loop)

        if self._model is None:
            logger.warning(
                "Creating a PlateModel instance with the 'model configuration' is None. Normally this should not happen. I allow this just in case some genius is doing something extraordinarily smart."
            )

    @property
    def model(self) -> Dict:
        """The model metadata."""
        if self._model is not None:
            return self._model
        else:
            raise Exception(
                "The 'model configuration' is None. This should not happen. Something extraordinary must have happened. Think carefully what you have done!!!"
            )

    @model.setter
    def model(self, var) -> None:
        if var is None:
            logger.warning(
                "You are trying to set the 'model configuration' to None. I will allow this. But I hope you know what you are doing!!!"
            )
        self._model = var

    def __getstate__(self):
        attributes = self.__dict__.copy()
        del attributes["executor"]
        del attributes["loop"]
        del attributes["run"]
        return attributes

    def __setstate__(self, state):
        self.__dict__ = state
        if not self.readonly:
            # async and concurrent things
            self.executor = concurrent.futures.ThreadPoolExecutor(max_workers=15)
            self.loop = asyncio.new_event_loop()
            self.run = functools.partial(self.loop.run_in_executor, self.executor)
            asyncio.set_event_loop(self.loop)

    def __del__(self):
        if not self.readonly:
            try:
                self.loop.close()
            except:
                pass  # ignore the exception when closing the loop if any

    def get_cfg(self):
        """Return the model configuration."""
        return self.model

    def get_model_dir(self):
        """Return the path to a folder containing the model files."""
        if PlateModel.is_model_dir(self.model_dir):
            return self.model_dir
        elif not self.readonly:
            return self.create_model_dir()
        else:
            raise Exception(
                f"The model dir {self.model_dir} is invalid and could not create it (in readonly mode)."
            )

    def get_data_dir(self):
        """Return the path to a folder (parent folder of the ``model dir``) containing a set of downloaded models."""
        return self.data_dir

    def set_data_dir(self, new_dir):
        """Change the folder (parent folder of the ``model dir``) in which you would like to save your model."""
        self.data_dir = new_dir
        self.model_dir = f"{self.data_dir}/{self.model_name}/"

    def get_big_time(self):
        """The max (big number in Ma) reconstruction time in the model."""
        return self.model["BigTime"]

    def get_small_time(self):
        """The min (small number in Ma) reconstruction time in the model."""
        return self.model["SmallTime"]

    def get_avail_layers(self):
        """Get all available layers in this plate model."""
        if not self.model:
            raise Exception("Fatal: No model configuration found!")
        return list(self.model["Layers"].keys())

    def get_rotation_model(self):
        """Return a list of rotation files."""
        if not self.readonly:
            rotation_folder = self._download_layer_files("Rotations")
        else:
            rotation_folder = f"{self.model_dir}/Rotations"
        rotation_files = glob.glob(f"{rotation_folder}/*.rot")
        rotation_files.extend(glob.glob(f"{rotation_folder}/*.grot"))
        # print(rotation_files)
        return rotation_files

    def get_coastlines(
        self, return_none_if_not_exist: bool = False
    ) -> Union[List[str], None]:
        """Return a list of ``coastlines`` files."""
        return self.get_layer(
            "Coastlines", return_none_if_not_exist=return_none_if_not_exist
        )

    def get_static_polygons(
        self, return_none_if_not_exist: bool = False
    ) -> Union[List[str], None]:
        """Return a list of ``static polygons`` files."""
        return self.get_layer(
            "StaticPolygons", return_none_if_not_exist=return_none_if_not_exist
        )

    def get_continental_polygons(
        self, return_none_if_not_exist: bool = False
    ) -> Union[List[str], None]:
        """Return a list of ``continental polygons`` files."""
        return self.get_layer(
            "ContinentalPolygons", return_none_if_not_exist=return_none_if_not_exist
        )

    def get_topologies(
        self, return_none_if_not_exist: bool = False
    ) -> Union[List[str], None]:
        """Return a list of ``topologies`` files."""
        return self.get_layer(
            "Topologies", return_none_if_not_exist=return_none_if_not_exist
        )

    def get_COBs(
        self, return_none_if_not_exist: bool = False
    ) -> Union[List[str], None]:
        """Return a list of ``Continent-Ocean Boundaries`` files."""
        return self.get_layer("COBs", return_none_if_not_exist=return_none_if_not_exist)

    def get_layer(
        self, layer_name: str, return_none_if_not_exist: bool = False
    ) -> Union[List[str], None]:
        """Get a list of layer files by a layer name. Call :meth:`get_avail_layers` to get all the available layer names.

        Raise :class:`LayerNotFoundInModel` exception to get user's attention by default.
        Set ``return_none_if_not_exist`` to ``True`` if you don't want to see the :class:`LayerNotFoundInModel` exception.

        :param layer_name: The layer name of interest.
        :param return_none_if_not_exist: If set to ``True``, return ``None`` when the layer does not exist in the model.

        :returns: A list of file names or ``None`` if ``return_none_if_not_exist`` is set to ``True``.

        :raises :class:`LayerNotFoundInModel`: Raise this exception if the layer name does not exist in this model.

        """
        try:
            if not self.readonly:
                layer_folder = self._download_layer_files(layer_name)
            else:
                layer_folder = f"{self.model_dir}/{layer_name}"
            files = []
            for ext in FILE_EXT:
                files.extend(glob.glob(f"{layer_folder}/*.{ext}"))

            return files
        except LayerNotFoundInModel as e:
            logger.warning(e)
            if return_none_if_not_exist:
                logger.warning(
                    f"The layer({layer_name}) does not exist in model({self.model_name})."
                )
                return None
            else:
                raise e

    def get_raster(self, raster_name: str, time: Union[int, float]) -> str:
        """Return a local path for the raster file.

        :param time: A single time of interest.
        :type time: int or float

        :returns: A local path of the raster file.
        :rtype: str
        """

        if not "TimeDepRasters" in self.model:
            raise Exception("No time-dependent rasters found in this model.")
        if not raster_name in self.model["TimeDepRasters"]:
            raise Exception(
                f"Time-dependent rasters ({raster_name}) not found in this model. {self.model['TimeDepRasters']}"
            )
        url = self.model["TimeDepRasters"][raster_name].format(time)

        if not self.readonly:
            self._download_raster(url, f"{self.get_model_dir()}/Rasters/{raster_name}")
        file_name = url.split("/")[-1]
        local_path = f"{self.get_model_dir()}/Rasters/{raster_name}/{file_name}"
        if os.path.isfile(local_path):
            return local_path
        elif self.readonly:
            raise Exception(
                f"You are in readonly mode and the raster {url} has not been downloaded yet."
            )
        else:
            raise Exception(f"Failed to download {url}")

    def get_rasters(
        self, raster_name: str, times: List[Union[int, float]]
    ) -> List[str]:
        """Return local paths for the raster files.

        :param times: A list of times
        :returns: A list of local paths
        """
        if not "TimeDepRasters" in self.model:
            raise Exception("No time-dependent rasters found in this model.")
        if not raster_name in self.model["TimeDepRasters"]:
            raise Exception(
                f"Time-dependent rasters ({raster_name}) not found in this model. {self.model['TimeDepRasters']}"
            )

        if not self.readonly:
            self.download_time_dependent_rasters(raster_name, times)

        paths = []
        for time in times:
            url = self.model["TimeDepRasters"][raster_name].format(time)
            file_name = url.split("/")[-1]
            local_path = f"{self.get_model_dir()}/Rasters/{raster_name}/{file_name}"
            if os.path.isfile(local_path):
                paths.append(local_path)
            elif self.readonly:
                raise Exception(
                    f"You are in readonly mode and the raster {url} has not been downloaded yet."
                )
            else:
                raise Exception(f"Failed to download {url}")
        return paths

    def create_model_dir(self):
        """Create a folder with a file ``.metadata.json`` in it to keep the model files."""
        if self.readonly:
            raise Exception("Unable to create model dir in readonly mode.")
        if not self.model_dir:
            raise Exception(f"Error: model dir is {self.model_dir}")

        # model dir already exists
        if PlateModel.is_model_dir(self.model_dir):
            return self.model_dir

        model_path = self.model_dir
        if os.path.isfile(model_path):
            raise Exception(
                f"Fatal: the model folder {model_path} already exists and is a file!! Remove the file or use another path."
            )

        Path(model_path).mkdir(parents=True, exist_ok=True)

        metadata_file = f"{model_path}/{self.meta_filename}"
        if not os.path.isfile(metadata_file):
            with open(metadata_file, "w+") as f:
                json.dump(self.model, f)

        return model_path

    @staticmethod
    def is_model_dir(folder_path: str):
        """Return ``True`` if the folder contains files of a plate model, otherwise ``False``."""
        return os.path.isdir(folder_path) and os.path.isfile(
            f"{folder_path}/.metadata.json"
        )

    def purge(self):
        """Remove the model folder and everything inside the folder."""
        if os.path.isdir(self.model_dir):
            shutil.rmtree(self.model_dir)

    def purge_layer(self, layer_name):
        """Remove the layer folder of the given layer name."""
        layer_path = f"{self.model_dir}/{layer_name}"
        if os.path.isdir(layer_path):
            shutil.rmtree(layer_path)

    def purge_time_dependent_rasters(self, raster_name):
        """Remove the raster folder of the given raster name."""
        raster_path = f"{self.model_dir}/{raster_name}"
        if os.path.isdir(raster_path):
            shutil.rmtree(raster_path)

    def _download_layer_files(self, layer_name):
        """Download layer files for a given layer name. You should use :meth:`get_layer`, instead of this one, whenever possible.

        The layer files are in a ".zip" file. This function will download and unzip it.

        :param layer_name: the layer name, such as "Rotations","Coastlines", "StaticPolygons", "ContinentalPolygons", "Topologies", etc.
                           Call :meth:`get_avail_layers` to get all the available layer names.

        :returns: the folder path which contains the layer files

        """
        if self.readonly:
            raise Exception("Unable to download layer files in readonly mode.")

        layer_file_url = self._get_layer_file_url(layer_name)

        model_folder = self.create_model_dir()
        layer_folder = f"{model_folder}/{layer_name}"
        metadata_file = f"{layer_folder}/{self.meta_filename}"

        downloader = download.FileDownloader(
            layer_file_url, metadata_file, model_folder, timeout=self.timeout
        )
        # only re-download when necessary
        if downloader.check_if_file_need_update():
            if os.path.isdir(layer_folder):
                # move the old layer files into "history" folder
                timestamp_str = datetime.now().strftime("%Y-%m-%d_%H-%M-%S")
                history_dir = f"{model_folder}/history/{layer_name}_{timestamp_str}"
                Path(history_dir).mkdir(parents=True, exist_ok=True)
                shutil.move(layer_folder, history_dir)

            downloader.download_file_and_update_metadata()
        else:
            if downloader.check_if_expire_date_need_update():
                # update the expiry date
                downloader.update_metadata()

            logger.debug(
                f"The local files in {layer_folder} are still good. Will not download again at this moment."
            )

        return layer_folder

    def download_all_layers(self):
        """Download all layers. This function calls :meth:`download_layer_files()` on every available layer."""
        if self.readonly:
            raise Exception("Unable to download all layers in readonly mode.")

        async def f():
            tasks = []
            if "Rotations" in self.model:
                tasks.append(self.run(self._download_layer_files, "Rotations"))
            if "Layers" in self.model:
                for layer in self.model["Layers"]:
                    tasks.append(self.run(self._download_layer_files, layer))

            # print(tasks)
            await asyncio.wait(tasks)

        try:
            self.loop.run_until_complete(f())
        except RuntimeError:
            import nest_asyncio

            nest_asyncio.apply()
            self.loop.run_until_complete(f())

    def get_avail_time_dependent_raster_names(self):
        """Return all time-dependent raster names in this plate model."""
        if not "TimeDepRasters" in self.model:
            return []
        else:
            return [name for name in self.model["TimeDepRasters"]]

    def download_time_dependent_rasters(self, raster_name, times=None):
        """Download time-dependent rasters for a given raster name.

        Call :meth:`get_avail_time_dependent_raster_names()` to see all the available raster names in this model.

        :param raster_name: the raster name of interest
        :param times: if not given, download from begin to end with 1My interval
        """
        if self.readonly:
            raise Exception(
                "Unable to download time dependent rasters in readonly mode."
            )

        if (
            "TimeDepRasters" in self.model
            and raster_name in self.model["TimeDepRasters"]
        ):

            async def f():
                nonlocal times
                tasks = []

                dst_path = f"{self.get_model_dir()}/Rasters/{raster_name}"
                if not times:
                    times = range(self.model["SmallTime"], self.model["BigTime"] + 1)
                for time in times:
                    tasks.append(
                        self.run(
                            self._download_raster,
                            self.model["TimeDepRasters"][raster_name].format(time),
                            dst_path,
                        )
                    )

                # print(tasks)
                await asyncio.wait(tasks)

            try:
                self.loop.run_until_complete(f())
            except RuntimeError:
                import nest_asyncio

                nest_asyncio.apply()
                self.loop.run_until_complete(f())

        else:
            raise Exception(
                f"Unable to find {raster_name} configuration in this model {self.model_name}."
            )

    def _download_raster(self, url, dst_path):
        """Download a single raster file from ``url`` and save the file in ``dst_path``.

        A metadata file will also be created for the raster file in folder ``f"{dst_path}/metadata"``

        :param url: the url to the raster file
        :param dst_path: the folder path to save the raster file

        """
        if self.readonly:
            raise Exception("Unable to download raster in readonly mode.")
        filename = url.split("/")[-1]
        metadata_folder = f"{dst_path}/.metadata"
        metadata_file = f"{metadata_folder}/{filename}.json"

        downloader = download.FileDownloader(
            url, metadata_file, dst_path, timeout=self.timeout
        )
        # only re-download when necessary
        if downloader.check_if_file_need_update():
            downloader.download_file_and_update_metadata()
        else:
            if downloader.check_if_expire_date_need_update():
                # update the expiry date
                downloader.update_metadata()

            logger.debug(
                f"The local raster file {dst_path}/{filename} is still good. Will not download again at this moment."
            )

    def download_all(self):
        """Download everything in this plate model."""
        if self.readonly:
            raise Exception("Unable to download all in readonly mode.")
        self.download_all_layers()
        if "TimeDepRasters" in self.model:
            for raster in self.model["TimeDepRasters"]:
                self.download_time_dependent_rasters(raster)

    def _get_layer_file_url(self, layer_name: str):
        # find the layer file url. two parts. one is the rotation, the other is all other geometry layers
        if layer_name == "Rotations":
            # for Rotations
            return self.model[layer_name]
        elif "Layers" in self.model and layer_name in self.model["Layers"]:
            # for other geometry layers
            return self.model["Layers"][layer_name]
        else:
            logger.debug(f"{json.dumps(self.model, indent=4)}")
            raise LayerNotFoundInModel(
                f"The layer({layer_name}) was not found in model({self.model_name})."
            )



import logging
import os
import sys
import warnings

pmm_logger = logging.getLogger("pmm")
stdout_handler = logging.StreamHandler(sys.stdout)
formatter = logging.Formatter(
    fmt="%(asctime)s - %(name)s - %(levelname)s - %(message)s",
    datefmt="%Y-%m-%d,%H:%M:%S",
)


def setup_logging():
    pmm_logger.propagate = False
    stdout_handler.setFormatter(formatter)
    stdout_handler.setLevel(logging.DEBUG)
    pmm_logger.addHandler(stdout_handler)
    pmm_logger.setLevel(logging.INFO)

    if is_debug_mode():
        turn_on_debug_logging()


def add_logging_file(filename: str = "pmm.log", level=logging.INFO):
    fh = logging.FileHandler(filename)
    fh.setLevel(level)
    fh.setFormatter(formatter)
    pmm_logger.addHandler(fh)


def turn_on_debug_logging():
    set_logging_level(logging.DEBUG)
    pmm_logger.debug(f"The debug logging has been enabled.")


def set_logging_level(level=logging.WARNING):
    logger = logging.getLogger("pmm")
    logger.setLevel(level)
    for h in logger.handlers:
        h.setLevel(level)


def disable_stdout_logging():
    pmm_logger.removeHandler(stdout_handler)
    print("The logging to stdout has been disabled.")


def my_warningformat(message, category, filename, lineno, line=None):
    return f"{filename}:{lineno}: {category.__name__}: {message}\n"


warnings.formatwarning = my_warningformat


def print_warning(msg):
    # warnings.warn(msg)
    pmm_logger.warning(msg)


def print_error(msg):
    pmm_logger.error(msg)


def is_debug_mode():
    return "PMM_DEBUG" in os.environ and os.environ["PMM_DEBUG"].lower() == "true"


from importlib.metadata import PackageNotFoundError, version


def get_distribution_version():
    """get the version string from the package metadata"""

    try:
        return version("plate_model_manager")
    except PackageNotFoundError:
        return "UNKNOWN VERSION"



from typing import Dict, List, Union

import requests


class ZenodoRecord:
    def __init__(self, conceptrecid):
        self.conceptrecid = conceptrecid
        self._record_url = f"https://zenodo.org/api/records?q=conceptrecid:{conceptrecid}&all_versions=true"
        r = requests.get(self._record_url)
        self.sub_records = r.json()["hits"]["hits"]

    def get_all_versions(self) -> List[Dict]:
        return self.sub_records

    def get_all_version_ids(self) -> List[str]:
        return [record["id"] for record in self.get_all_versions()]

    def get_version(self, id: int) -> Dict:
        for record in self.get_all_versions():
            if record["id"] == id:
                return record

        raise Exception(
            f"Unable to get version({id}). Check {self._record_url} to find out what is going on."
        )

    def get_latest_version(self) -> Dict:
        for record in self.sub_records:
            if record["metadata"]["relations"]["version"][0]["is_last"] == True:
                return record

        raise Exception(
            f"Unable to find the latest version. Check {self._record_url} to find out what is going on."
        )

    def get_latest_version_id(self) -> int:
        return self.get_latest_version()["id"]

    def get_file_links(self, id: int) -> List[str]:
        record = self.get_version(id)
        return [file["links"]["self"] for file in record["files"]]

    def get_filenames(self, id: int) -> List[str]:
        record = self.get_version(id)
        return [file["key"] for file in record["files"]]



import json
import logging
import os
import re
from typing import Dict, Union

import requests

from .exceptions import InvalidConfigFile, ServerUnavailable
from .plate_model import PlateModel

logger = logging.getLogger("pmm")


class PlateModelManager:
    """Manage a set of publicly available plate reconstruction models.
    The model files are hosted on EarthByte servers.
    You need Internet connection to use this class and download the files.
    """

    # Load a models.json file and manage plate models.
    # See an example models.json file at PlateModelManager.get_default_repo_url().

    def __init__(self, model_manifest: str = "", timeout=(None, None)):
        """Constructor. Create a :class:`PlateModelManager` instance.
        You need Internet connection to create an instance of this class.
        If you don't have Internet connection, use :class:`PlateModel` class directly in ``readonly`` mode.
        Visit `this page <examples.html#use-without-internet>`__ to see an example.

        :param model_manifest: The URL to a ``models.json`` metadata file.
                               Normally you don't need to provide this parameter unless
                               you would like to setup your own plate model server.

        """
        if not model_manifest:
            self.model_manifest = PlateModelManager.get_default_repo_url()
        else:
            self.model_manifest = model_manifest

        self._models = None
        self.timeout = timeout

        if not isinstance(self.model_manifest, str):
            raise InvalidConfigFile(
                f"The model_manifest '{type(self.model_manifest)}' must be a string. It is either a local file path or a http(s) URL."
            )

        # check if the model manifest file is a local file
        if os.path.isfile(self.model_manifest):
            with open(self.model_manifest) as f:
                self._models = json.load(f)
        elif self.model_manifest.startswith(
            "http://"
        ) or self.model_manifest.startswith("https://"):
            # try the http(s) url
            try:
                r = requests.get(self.model_manifest, timeout=timeout)
                if r.status_code != 200:
                    raise InvalidConfigFile(
                        f"Unable to get valid JSON data from '{self.model_manifest}'. Http request return code: {r.status_code}"
                    )
                else:
                    self._models = r.json()

            except (
                requests.exceptions.ConnectionError,
                requests.exceptions.ConnectTimeout,
                requests.exceptions.ReadTimeout,
            ):
                raise ServerUnavailable(
                    f"Unable to fetch {self.model_manifest}. No network connection, server unavailable or invalid URL!"
                )
            except requests.exceptions.JSONDecodeError:
                raise InvalidConfigFile(
                    f"Unable to get valid JSON data from '{self.model_manifest}'."
                )
        else:
            raise InvalidConfigFile(
                f"The model_manifest '{self.model_manifest}' must be either a local file path or a http(s) URL."
            )

        if "vars" in self.models:
            self._replace_vars_with_values(self.models["vars"], self.models)
            del self.models["vars"]

    @property
    def models(self) -> Dict:
        """The metadata for all the models."""
        if self._models is not None:
            return self._models
        else:
            raise Exception(
                f"No model found. Check the model manifest {self.model_manifest} for errors."
            )

    @models.setter
    def models(self, var) -> None:
        self._models = var

    def _replace_vars_with_values(self, var_dict, json_obj):
        """Replace the variables in `json_obj` with the real values. The variables are defined in `var_dict`."""
        for key, value in json_obj.items():
            if key == "vars":
                continue
            if isinstance(value, dict):
                self._replace_vars_with_values(var_dict, value)
            elif isinstance(value, str):
                matches = re.findall("@<<(.*)>>@", value)
                for m in matches:
                    if m in var_dict:
                        value = value.replace(f"@<<{m}>>@", var_dict[m])
                json_obj[key] = value
            else:
                continue

    def get_model(
        self, model_name: str = "default", data_dir: str = "."
    ) -> Union[PlateModel, None]:
        """Return a :class:`PlateModel` object for a given plate model name.

        Call :meth:`get_available_model_names()` to see a list of available plate model names.

        :param model_name: the plate model name of interest
        :param data_dir: The folder to save the plate model files.
                         This ``data_dir`` can be changed with :meth:`PlateModel.set_data_dir()` later.

        :returns: a :class:`PlateModel` object or ``None`` if the plate model name is no good.

        """
        model_name = model_name.lower()
        if model_name in self.models:
            # model name is an alias
            if isinstance(self.models[model_name], str):
                m_name = self.models[model_name]
                if m_name.startswith("@"):
                    m_name = self.models[model_name][1:]

                m = self.get_model(m_name, data_dir=data_dir)
                if m is None:
                    raise Exception(
                        f"Unable to find model {m_name} to resolve an alias. There must be errors in the {self.model_manifest}"
                    )
                else:
                    return PlateModel(
                        model_name, model_cfg=m.get_cfg(), data_dir=data_dir
                    )
            else:
                return PlateModel(
                    model_name, model_cfg=self.models[model_name], data_dir=data_dir
                )
        else:
            logger.error(f"Model {model_name} is not available.")
            return None

    def get_available_model_names(self):
        """Return the names of available models as a list."""
        return list(self.models.keys())

    @staticmethod
    def get_local_available_model_names(local_dir: str):
        """Return a list of model names in a local folder.

        :param local_dir: The local folder containing models.
        :type local_dir: str
        """
        models = []
        for file in os.listdir(local_dir):
            d = os.path.join(local_dir, file)
            if os.path.isdir(d) and os.path.isfile(f"{d}/.metadata.json"):
                models.append(file)
        return models

    @staticmethod
    def get_default_repo_url():
        """Return the URL to the configuration data of models."""
        default_repo_url_list = [
            "https://repo.gplates.org/webdav/pmm/config/models_v2.json",
            "https://www.earthbyte.org/webdav/pmm/config/models_v2_eb.json",
            "https://portal.gplates.org/static/pmm/config/models_v2_gp.json",
        ]
        for url in default_repo_url_list:
            try:
                response = requests.head(url, timeout=(5, 5))
                if response.status_code == 200:
                    return url
                else:
                    logger.warning(
                        f"Unable to fetch {url}. status_code={response.status_code}"
                    )
                    continue
            except:
                logger.warning(f"Unable to fetch {url}.")
                continue
        raise ServerUnavailable(
            """Cannot connect to the servers. Either the servers are currently unavailable, or there is a problem with your internet connection."""
        )

    def download_all_models(self, data_dir: str = "./") -> None:
        """Download all available models into the ``data_dir``.

        :param data_dir: The folder to save the model files.
        :type data_dir: str
        """
        for name in self.get_available_model_names():
            print(f"download {name}")
            model = self.get_model(name)
            if model is not None:
                model.set_data_dir(data_dir)
                model.download_all_layers()



import logging
import os
import re
from typing import Union

from .exceptions import ServerUnavailable
from .plate_model import PlateModel
from .plate_model_manager import PlateModelManager
from .zenodo import ZenodoRecord

logger = logging.getLogger("pmm")


def get_plate_model(
    model_name: str, data_dir: Union[str, os.PathLike]
) -> Union[PlateModel, None]:
    """Convenient function to create a :class:`PlateModel` instance using best effort.
    First, try to get the plate model with :class:`PlateModelManager`.
    If the servers cannot be reached, try to use the local plate model files which were previously downloaded.

    :param model_name: the plate model name of interest
    :param data_dir: The folder to save the plate model files.

    :returns: a :class:`PlateModel` object or ``None`` if the plate model name is no good.
    """
    try:
        model = PlateModelManager().get_model(model_name, data_dir=data_dir)
    except ServerUnavailable:
        # if unable to connect to the servers, try to use the local files
        model = PlateModel(model_name=model_name, data_dir=data_dir, readonly=True)
        logger.warning(
            "Unable to connect to the servers. Using local files in readonly mode."
        )
    return model


def check_update():
    """Check if new versions of plate models are available on Zenodo.
    Mainly used by Michael Chin to update the PMM server.
    """
    need_update = False
    models = PlateModelManager().models
    for model_name in models:
        logger.info(f"Checking update for model -- {model_name} ...")
        model = models[model_name]
        if isinstance(model, dict) and "URL" in model and "Version" in model:
            record_id = re.findall(r"zenodo.(\d+)", model["URL"])
            version_id = re.findall(r"zenodo.(\d+)", model["Version"])
            if len(record_id) == 1 and len(version_id) == 1:
                # logger.info(record_id[0])
                latest_id = str(ZenodoRecord(record_id[0]).get_latest_version_id())
                if version_id[0] != latest_id:
                    need_update = True
                    logger.info(
                        f"Model ({model_name}) needs update. The latest version ID is: {latest_id}. Your current version ID is : {version_id[0]}."
                    )
    if not need_update:
        logger.info("All models are up-to-date.")



import abc
import asyncio
import io
import os
from pathlib import Path
from typing import Union

from .utils import misc, network, unzip


class FileFetcher(metaclass=abc.ABCMeta):
    @classmethod
    def __subclasshook__(cls, subclass):
        return (
            hasattr(subclass, "fetch_file")
            and callable(subclass.fetch_file)
            and hasattr(subclass, "fetch_files")
            and callable(subclass.fetch_files)
            or hasattr(subclass, "fetch_large_file")
            and callable(subclass.fetch_large_file)
            or NotImplemented
        )

    @abc.abstractmethod
    def fetch_file(
        self,
        url: str,
        filepath: str,
        filename: Union[str, None] = None,
        etag: Union[str, None] = None,
        auto_unzip: bool = True,
    ):
        """download a file from "url" and save to "filepath"
        You can give a new "filename" for the file.
        If "etag" is given, check if etag has changed. If not changed, do not download again.

        :param url: the url to download file from
        :param filepath: location to keep the file
        :param filename: new file name (optional)
        :param etag: old etag. if the old etag is the same with the one on server, do not download again.
        :param auto_unzip: bool flag to indicate if unzip .zip file automatically

        :returns: new etag

        """

        raise NotImplementedError

    @abc.abstractmethod
    def fetch_files(
        self,
        urls,
        filepaths,
        filenames=[],
        etags=[],
        auto_unzip: bool = True,
    ):
        """fetch multiple files concurrently

        :param urls: the urls to download files from
        :param filepaths: location(s) to keep the files. This can be one path for all files or one path for each file.
        :param filenames: new file names (optional)
        :param etags: old etags. if the old etag is the same with the one on server, do not download again.
        :param auto_unzip: bool flag to indicate if unzip .zip file automatically

        """
        raise NotImplementedError

    @abc.abstractmethod
    def _run_fetch_large_file(self, loop, url, filesize, data):
        raise NotImplementedError

    def fetch_large_file(
        self,
        url: str,
        filepath: str,
        filename: Union[str, None],
        filesize: Union[int, None] = None,
        etag: Union[str, None] = None,
        auto_unzip: bool = True,
        check_etag: bool = True,
        timeout=(None, None),
    ):
        """use multi-thread to fetch a large file.
            LOOK HERE!!!
            Be careful when use this function. You cannot get partial content if the content is gzip encoded.
            So the file might be larger than the one download directly.
            It is useful when downloading large .zip file.
            Warning: this could be slower than single thread download.
            Some firewall sequences these requests to shape network traffic and defeat the purpose of this function completely.

            check the etag and get content-length before calling this function

        :param url: the file url
        :param filepath: location to keep the file
        :param filename: the output filename specified by caller. ignored if auto unzip compressed data.
        :param filesize: the size of file (in bytes)
        :param etag: old etag. if the old etag is the same with the one on server, do not download again.
        :param auto_unzip: bool flag to indicate if unzip .zip file automatically

        :returns: new etag

        """
        if filesize is None or check_etag:
            headers = network.get_headers(url, timeout=timeout)
            file_size = network.get_content_length(headers)
            new_etag = network.get_etag(headers)
            # if the etags are the same, do not download again
            if etag and etag == new_etag:
                return etag
        else:
            # if the caller has provided file size and checked etag
            file_size = filesize
            new_etag = etag

        # create folder to keep the file
        if os.path.isfile(filepath):
            raise Exception(
                f"The 'filepath' is in fact a file. The 'filepath' should be a folder path(non-exist is fine). {filepath}"
            )
        Path(filepath).mkdir(parents=True, exist_ok=True)

        # set up concurrent functions
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)

        data = [io.BytesIO()]

        try:
            self._run_fetch_large_file(loop, url, file_size, data)

        except RuntimeError:
            import nest_asyncio

            nest_asyncio.apply()
            self._run_fetch_large_file(loop, url, file_size, data)
        except Exception as e:
            misc.print_error("Failed to fetch large file!")
            raise Exception("Failed to fetch large file!") from e
        finally:
            loop.close()

        data[0].seek(0)

        if not filename:
            filename = url.split("/")[-1]  # use the filename in the url
        # save the file
        if auto_unzip:
            try:
                unzip.save_compressed_data(url, data[0], filepath)
            except Exception as ex:
                # print(ex)
                # print("failed to save zip. try save directly")
                data[0].seek(0)
                self._save_file(filepath, filename, data[0].read())
        else:
            self._save_file(filepath, filename, data[0].read())

        return new_etag

    def _save_file(self, filepath, filename, data):
        """helper function to save file to hard drive"""

        Path(filepath).mkdir(parents=True, exist_ok=True)
        if os.path.isfile(f"{filepath}/{filename}"):
            print(f"Warning: overwriting {filename}")
        with open(f"{filepath}/{filename}", "wb+") as of:
            of.write(data)



import asyncio
import io
import os
from pathlib import Path
from typing import List, Union

import aiohttp

from .file_fetcher import FileFetcher
from .utils import unzip

# This file contains experimental code to download files concurrently using aiohttp.
# Later, I realized that "requests"+"ThreadPoolExecutor" works as well.
# I do not want to introduce a new dependency when "requests" works.
# So, keep this file just for record keeping in case we need aiohttp in the future.


class AiohttpFetcher(FileFetcher):
    def __init__(self):
        pass

    def fetch_file(
        self,
        url: str,
        filepath: str,
        filename: Union[str, None] = None,
        etag: Union[str, None] = None,
        auto_unzip: bool = True,
    ):
        """download a file from "url" and save to "filepath"
            You can give a new "filename" for the file.
            If "etag" is given, check if etag has changed. If not changed, do not download again.

        :param url: the url to download file from
        :param filepath: location to keep the file
        :param etag: old etag. if the old etag is the same with the one on server, do not download again.
        :param auto_unzip: bool flag to indicate if unzip .zip file automatically

        """

        async def f():
            async with aiohttp.ClientSession() as session:
                await self._fetch_file(
                    session,
                    url,
                    filepath,
                    filename=filename,
                    etag=etag,
                    auto_unzip=auto_unzip,
                )
            await asyncio.sleep(
                0.250
            )  # https://docs.aiohttp.org/en/stable/client_advanced.html#graceful-shutdown

        asyncio.run(f())

    async def _fetch_file(
        self,
        session,
        url: str,
        filepath: str,
        filename: Union[str, None] = None,
        etag: Union[str, None] = None,
        auto_unzip: bool = True,
    ):
        """async "fetch_file" implementation. See the docstring of "fetch_file" """

        if isinstance(etag, str) or isinstance(etag, bytes):
            headers = {"If-None-Match": etag}
        else:
            headers = {}

        if os.path.isfile(filepath):
            raise Exception(
                f"The 'filepath' is in fact a file. The 'filepath' should be a folder path(non-exist is fine). {filepath}"
            )
        Path(filepath).mkdir(parents=True, exist_ok=True)

        async with session.get(url, headers=headers) as r:
            content = await r.content.read()
            # r = requests.get(url, allow_redirects=True, headers=headers)
            # print(r.headers)

            if r.status == 304:
                # print(url)
                print(
                    "The file has not been changed since it was downloaded last time. Do nothing and return."
                )
            elif r.status == 200:
                if not filename:
                    filename = url.split("/")[-1]  # use the filename in the url
                if auto_unzip:
                    try:
                        unzip.save_compressed_data(url, io.BytesIO(content), filepath)
                    except Exception as ex:
                        # print(ex)
                        self._save_file(filepath, filename, content)
                else:
                    self._save_file(filepath, filename, content)
            else:
                raise Exception(f"HTTP request failed with code {r.status_code}.")
            new_etag = r.headers.get("ETag")
            if new_etag:
                # remove the content-encoding awareness thing
                new_etag = new_etag.replace("-gzip", "")

            return new_etag

    async def _fetch_range(
        self, session, url: str, index: int, chunk_size: int, data: List
    ):
        """async funtion to get patial content of a file from the server
        Be careful, some server does not support this function.
        And some firewall sequences these requests to shape network traffic and defeat the purpose
        of this function completely.

        """
        # print(index)
        # st = time.time()
        headers = {
            "Range": f"bytes={index*chunk_size}-{(index+1)*chunk_size-1}",
            "Accept-Encoding": "identity",
        }

        # r = requests.get(url, headers=headers)
        async with session.get(url, headers=headers) as r:
            if r.status == 206:
                c = await r.content.read()
                data[index].write(c)
            else:
                raise Exception(f"Failed to fetch range from {url} at index {index}")
        # et = time.time()
        # print(f"{index} -- time: {et - st}")

    async def _fetch_large_file(
        self, url: str, file_size: int, data: List, chunk_size=10 * 1000 * 1000
    ):
        """async implementation of fetch_large_file"""
        async with aiohttp.ClientSession() as session:
            num_chunks = file_size // chunk_size + 1
            data_array = [io.BytesIO() for i in range(num_chunks)]
            tasks = [
                asyncio.ensure_future(
                    self._fetch_range(session, url, i, chunk_size, data_array)
                )
                for i in range(num_chunks)
            ]

            await asyncio.wait(tasks)

            for i in range(num_chunks):
                data_array[i].seek(0)
                data[0].write(data_array[i].read())

    def _run_fetch_large_file(self, loop, url, filesize, data):
        """run async function"""
        loop.run_until_complete(self._fetch_large_file(url, filesize, data))

    def fetch_files(
        self,
        urls,
        filepaths,
        filenames=[],
        etags=[],
        auto_unzip: bool = True,
    ):
        """fetch multiple files concurrently

        :param urls: the urls to download files from
        :param filepaths: location(s) to keep the files. This can be one path for all files or one path for each file.
        :param etags: old etags. if the old etag is the same with the one on server, do not download again.
        :param auto_unzip: bool flag to indicate if unzip .zip file automatically

        """

        async def f():
            async with aiohttp.ClientSession() as session:
                tasks = []
                for idx, url in enumerate(urls):
                    # get filepath
                    if isinstance(filepaths, str):
                        filepath = filepaths
                    elif isinstance(filepaths, list) and len(filepaths) > idx:
                        filepath = filepaths[idx]
                    else:
                        raise Exception(
                            "The 'filepaths' should be either one string or a list of strings. And the length of the list should be the same with the length of urls. "
                        )

                    # get etag
                    if len(etags) > idx:
                        etag = etags[idx]
                    else:
                        etag = None

                    tasks.append(
                        asyncio.ensure_future(
                            self._fetch_file(
                                session,
                                url,
                                filepath,
                                None,
                                etag=etag,
                                auto_unzip=auto_unzip,
                            )
                        )
                    )

                return await asyncio.gather(*tasks)

        # set up concurrent functions
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        new_etags = []
        try:
            new_etags = loop.run_until_complete(f())
        except RuntimeError:
            import nest_asyncio

            nest_asyncio.apply()
            new_etags = loop.run_until_complete(f())
        finally:
            loop.close()
            return new_etags


def fetch_file(
    url: str,
    filepath: str,
    filename: Union[str, None] = None,
    etag: Union[str, None] = None,
    auto_unzip: bool = True,
):
    fetcher = AiohttpFetcher()
    return fetcher.fetch_file(
        url, filepath, filename=filename, etag=etag, auto_unzip=auto_unzip
    )


def fetch_files(
    urls,
    filepaths,
    etags=[],
    auto_unzip: bool = True,
):
    fetcher = AiohttpFetcher()
    return fetcher.fetch_files(urls, filepaths, etags=etags, auto_unzip=auto_unzip)


def fetch_large_file(
    url: str,
    filepath: str,
    filename: Union[str, None] = None,
    filesize: Union[int, None] = None,
    etag: Union[str, None] = None,
    auto_unzip: bool = True,
    check_etag: bool = True,
):
    fetcher = AiohttpFetcher()
    return fetcher.fetch_large_file(
        url,
        filepath,
        filename=filename,
        filesize=filesize,
        etag=etag,
        auto_unzip=auto_unzip,
        check_etag=check_etag,
    )



import json
import logging
import os
from datetime import datetime, timedelta
from pathlib import Path
from typing import Union

from .. import network_aiohttp, network_requests
from . import network

EXPIRY_TIME_FORMAT = "%Y/%m/%d, %H:%M:%S"
EXPIRE_HOURS = 12

logger = logging.getLogger("pmm")

# {url:{new-etag:"xxxx", file-size:12345, meta-etag:"uuuuu"}}
etag_and_file_size_cache = {}

from enum import Enum


class HttpClient(Enum):
    REQUESTS = 1
    AIOHTTP = 2


class FileDownloader:
    """class for managing single file download"""

    def __init__(
        self,
        file_url: str,
        meta_filepath: str,
        dst_dir: str,
        filename: Union[str, None] = None,
        auto_unzip: bool = True,
        expire_hours=EXPIRE_HOURS,
        expiry_time_format=EXPIRY_TIME_FORMAT,
        large_file_hint=False,
        timeout=(None, None),
        http_client: HttpClient = HttpClient.REQUESTS,
    ) -> None:
        """FileDownloader constructor

        :param file_url: the url to the file
        :param metadata_file: the path to the metadata file
        :param dst_dir: the destination to save the file
        """
        self.file_url = file_url
        self.meta_filepath = meta_filepath
        self.dst_dir = dst_dir
        self.filename = filename
        self.expire_hours = expire_hours
        self.expiry_time_format = expiry_time_format
        self.meta_etag = None
        self.new_etag = None
        self.file_size = None
        self.large_file_hint = large_file_hint
        self.timeout = timeout
        self.auto_unzip = auto_unzip
        self.http_client = http_client

    def check_if_file_need_update(self):
        """check if the file need an update(download/re-download the files)
        return true if "need update", otherwise false

        1. check if the metadata file exists
        2. check if the file urls match
        3. check expire date
        4. check etag
        """

        #
        # first check if the metadata file exists
        # since metadata file is inside the layer folder, this check will also confirm the existence of the layer folder
        #
        if not os.path.isfile(self.meta_filepath):
            logger.debug(
                f"the metadata file({self.meta_filepath}) does not exist, need to download the file({self.file_url})"
            )
            return True

        with open(self.meta_filepath, "r") as f:
            meta = json.load(f)
            #
            # check if the "url" in the metafile matches the "layer file url"
            #
            if "url" in meta:
                meta_url = meta["url"]
                if meta_url != self.file_url:
                    logger.debug(
                        "the layer url has changed, re-download the file({file_url})"
                    )
                    return True
            else:
                logger.debug(
                    "no url found in the metafile. to be on the safe side, re-download the file({file_url})"
                )
                return True
            #
            # now check the layer file's expiry date
            #
            need_check_etag = False
            if "expiry" in meta:
                try:
                    meta_expiry = meta["expiry"]
                    expiry_date = datetime.strptime(
                        meta_expiry, self.expiry_time_format
                    )
                    now = datetime.now()
                    if now > expiry_date:
                        logger.debug("The file expired. Check etag.")
                        need_check_etag = True  # expired, need to check etag to decide
                    else:
                        # layer file has not expired yet, no need to check update
                        return False
                except ValueError:
                    need_check_etag = (
                        True  # invalid expiry date, need to check etag to decide
                    )
            else:
                need_check_etag = (
                    True  # no expiry date in metafile, need to check etag to make sure
                )

            if need_check_etag:
                if "etag" in meta:
                    meta_etag = meta["etag"]
                    headers = network.get_headers(self.file_url)
                    self.file_size = network.get_content_length(headers)
                    self.new_etag = network.get_etag(headers)

                    if meta_etag == self.new_etag:
                        logger.debug(f"{meta_etag} -- {self.new_etag}")
                        return False
                    else:
                        logger.debug(
                            f"etag has been changed. re-download the file({self.file_url})"
                        )
                        return True

                else:
                    logger.debug(
                        f"no etag found in the metadata file, to be safe, re-download the file({self.file_url})"
                    )
                    return True

            logger.debug("This line and below should not be reached!!!!")
            return True

    def download_file_and_update_metadata(self):
        """download a file from "file_url", save the file in "dst_dir" and update the metadata file

        :param file_url: the url to the file
        :param metadata_file: the path to the metadata file
        :param dst_dir: the destination to save the file

        """
        if self.large_file_hint:
            headers = network.get_headers(self.file_url)
            self.file_size = network.get_content_length(headers)

        if self.http_client == HttpClient.REQUESTS:
            client = network_requests
        else:
            client = network_aiohttp

        if self.file_size and self.file_size > 20 * 1000 * 1000:
            self.new_etag = client.fetch_large_file(
                self.file_url,
                self.dst_dir,
                filename=self.filename,
                filesize=self.file_size,
                etag=None,
                auto_unzip=self.auto_unzip,
                check_etag=False,
            )

        else:
            self.new_etag = client.fetch_file(
                self.file_url,
                self.dst_dir,
                filename=self.filename,
                etag=self.meta_etag,
                auto_unzip=self.auto_unzip,
            )

        # update metadata file
        self.update_metadata()

    def update_metadata(self):
        """update metadata file"""
        metadata = {
            "url": self.file_url,
            "expiry": (datetime.now() + timedelta(hours=self.expire_hours)).strftime(
                self.expiry_time_format
            ),
            "etag": self.new_etag,
        }
        Path("/".join(self.meta_filepath.split("/")[:-1])).mkdir(
            parents=True, exist_ok=True
        )
        with open(self.meta_filepath, "w+") as f:
            json.dump(metadata, f)

    def check_if_expire_date_need_update(self):
        # if we have checked the etag and it is the same as before
        # we need to update the expiry date
        return self.new_etag is not None and self.new_etag == self.meta_etag



# import functions into __init__.py

from .utils.misc import get_distribution_version

# __version__ = "1.3.0"
__version__ = get_distribution_version()
del get_distribution_version

from .auxiliary import check_update, get_plate_model
from .plate_model import PlateModel
from .plate_model_manager import PlateModelManager
from .present_day_rasters import PresentDayRasterManager
from .utils.misc import (
    disable_stdout_logging,
    is_debug_mode,
    set_logging_level,
    setup_logging,
    turn_on_debug_logging,
)

setup_logging()

__all__ = [
    "PlateModelManager",
    "PresentDayRasterManager",
    "PlateModel",
    "get_plate_model",
]



import bz2
import gzip
import lzma
import os
import shutil
import sys
import tarfile
import zipfile

from . import misc


def save_compressed_data(url, data, dst_path):
    """extract files from compressed data

    :param url: URL
    :param data: bytes-like object
    :param dst_path: location to save the files
    """
    # print(f"save_compressed_data:{url}")
    # .zip
    if url.endswith(".zip"):
        if not zipfile.is_zipfile(data):
            misc.print_warning(
                f"The {url} seems a zip file. But it is in fact not. Will not decompress the file."
            )
            raise Exception("Bad compressed data!")
        else:
            with zipfile.ZipFile(data) as z:
                z.extractall(dst_path)
                z.close()
    # .tar.gz or .tgz
    elif url.endswith(".tar.gz") or url.endswith(".tgz"):
        if sys.version_info[1] > 8 and (not tarfile.is_tarfile(data)):
            misc.print_warning(
                f"The {url} seems a tar gzip file. But it is in fact not. Will not decompress the file."
            )
            raise Exception("Bad compressed data!")
        else:
            data.seek(0)
            with tarfile.open(fileobj=data, mode="r:gz") as tar:
                tar.extractall(path=dst_path)
                tar.close()
    # .gz
    elif url.endswith(".gz"):
        fn = url.split("/")[-1][:-3]
        if not fn:
            raise Exception(
                f"The url ends with .gz. But unable to extract the file name. Check the URL {url}"
            )
        try:
            with open(f"{dst_path}/{fn}", "wb+") as f_out:
                with gzip.GzipFile(fileobj=data) as f_in:
                    shutil.copyfileobj(f_in, f_out)
                    f_in.close()
                f_out.close()
        except:
            os.remove(f"{dst_path}/{fn}")
            misc.print_warning(
                f"The {url} seems a gzip file. But it is in fact not. Will not decompress the file."
            )
            raise Exception("Bad compressed data!")
    # .tar.bz2 or .tbz2
    elif url.endswith(".tar.bz2") or url.endswith(".tbz2"):
        if sys.version_info[1] > 8 and (not tarfile.is_tarfile(data)):
            misc.print_warning(
                f"The {url} seems a tar bz2 file. But it is in fact not. Will not decompress the file."
            )
            raise Exception("Bad compressed data!")
        else:
            data.seek(0)
            with tarfile.open(fileobj=data, mode="r:bz2") as tar:
                tar.extractall(path=dst_path)
                tar.close()
    # .bz2
    elif url.endswith(".bz2"):
        fn = url.split("/")[-1][:-4]
        if not fn:
            raise Exception(
                f"The url ends with .bz2. But unable to extract the file name. Check the URL {url}"
            )
        try:
            with open(f"{dst_path}/{fn}", "wb+") as f_out:
                data = bz2.decompress(data.read())
                f_out.write(data)
                f_out.close()
        except:
            os.remove(f"{dst_path}/{fn}")
            misc.print_warning(
                f"The {url} seems a bz2 file. But it is in fact not. Will not decompress the file."
            )
            raise Exception("Bad compressed data!")
    # .lzma
    elif url.endswith(".lzma"):
        fn = url.split("/")[-1][:-5]
        if not fn:
            raise Exception(
                f"The url ends with .lzma. But unable to extract the file name. Check the URL {url}"
            )
        try:
            with open(f"{dst_path}/{fn}", "wb+") as f_out:
                data = lzma.decompress(data.read())
                f_out.write(data)
                f_out.close()
        except:
            os.remove(f"{dst_path}/{fn}")
            misc.print_warning(
                f"The {url} seems a lzma file. But it is in fact not. Will not decompress the file."
            )
            raise Exception("Bad compressed data!")
    # .tar.xz or .txz
    elif url.endswith(".tar.xz") or url.endswith(".txz"):
        if sys.version_info[1] > 8 and (not tarfile.is_tarfile(data)):
            misc.print_warning(
                f"The {url} seems a tar xz file. But it is in fact not. Will not decompress the file."
            )
            raise Exception("Bad compressed data!")
        else:
            data.seek(0)
            with tarfile.open(fileobj=data, mode="r:xz") as tar:
                tar.extractall(path=dst_path)
                tar.close()
    # .xz
    elif url.endswith(".xz"):
        fn = url.split("/")[-1][:-3]
        if not fn:
            raise Exception(
                f"The url ends with .xz. But unable to extract the file name. Check the URL {url}"
            )
        try:
            with open(f"{dst_path}/{fn}", "wb+") as f_out:
                data = lzma.decompress(data.read())
                f_out.write(data)
                f_out.close()
        except:
            os.remove(f"{dst_path}/{fn}")
            misc.print_warning(
                f"The {url} seems a xz file. But it is in fact not. Will not decompress the file."
            )
            raise Exception("Bad compressed data!")
    else:
        raise Exception("Unrecognized zip data!")



import asyncio
import concurrent.futures
import functools
import io
import os
from pathlib import Path
from typing import List, Union

import requests

from .file_fetcher import FileFetcher
from .utils import unzip

# This file contains the code to download file(s) from url(s) using "requests" + "asyncio"
# Download files concurrently can improve the performance significantly.
# Download partial content of a large file concurrently may/maynot work depending on the server configuration.


class RequestsFetcher(FileFetcher):
    def __init__(self):
        pass

    def fetch_file(
        self,
        url: str,
        filepath: str,
        filename: Union[str, None] = None,
        etag: Union[str, None] = None,
        auto_unzip: bool = True,
        timeout=(None, None),
    ):
        """download a file from "url" and save to "filepath"
            You can give a new "filename" for the file.
            If "etag" is given, check if etag has changed. If not changed, do not download again.

        :param url: the url to download file from
        :param filepath: location to keep the file
        :param etag: old etag. if the old etag is the same with the one on server, do not download again.
        :param auto_unzip: bool flag to indicate if unzip .zip file automatically

        """
        # print(f"url: {url}")
        # print(f"filepath: {filepath}")
        # print(f"filename: {filename}")
        # print(f"etag: {etag}")
        # print(f"auto_unzip: {auto_unzip}")
        # print(f"timeout: {timeout}")

        if isinstance(etag, str) or isinstance(etag, bytes):
            headers = {"If-None-Match": etag}
        else:
            headers = {}

        if os.path.isfile(filepath):
            raise Exception(
                f"The 'filepath' is in fact a file. The 'filepath' should be a folder path(non-exist is fine). {filepath}"
            )
        Path(filepath).mkdir(parents=True, exist_ok=True)

        r = requests.get(url, allow_redirects=True, headers=headers, timeout=timeout)
        # print(r.headers)

        if r.status_code == 304:
            # print(url)
            print(
                "The file has not been changed since it was downloaded last time. Do nothing and return."
            )
        elif r.status_code == 200:
            if filename is None:
                filename = url.split("/")[-1]  # use the filename in the url
            if auto_unzip:
                try:
                    unzip.save_compressed_data(url, io.BytesIO(r.content), filepath)
                except Exception as ex:
                    # print(ex)
                    self._save_file(filepath, filename, r.content)
            else:
                self._save_file(filepath, filename, r.content)
        else:
            raise Exception(f"HTTP request failed with code {r.status_code}. {url}")
        new_etag = r.headers.get("ETag")
        if new_etag:
            # remove the content-encoding awareness thing
            new_etag = new_etag.replace("-gzip", "")

        return new_etag

    def _fetch_range(
        self, url, index: int, chunk_size: int, data: List, timeout=(None, None)
    ):
        """get patial content of a file from the server
        Be careful, some server does not support this function.
        And some firewall sequences these requests to shape network traffic and defeat the purpose
        of this function completely. So it might slower than download directly.

        """
        # print(index)
        # st = time.time()
        headers = {
            "Range": f"bytes={index*chunk_size}-{(index+1)*chunk_size-1}",
            "Accept-Encoding": "identity",
        }

        r = requests.get(url, headers=headers, timeout=timeout)
        if r.status_code == 206:
            data[index].write(r.content)
        else:
            raise Exception(f"Failed to fetch range from {url} at index {index}")
        # et = time.time()
        # print(f"{index} -- time: {et - st}")

    def _run_fetch_large_file(self, loop, url, filesize, data, timeout=(None, None)):
        """run async function"""
        executor = concurrent.futures.ThreadPoolExecutor(max_workers=15)
        run = functools.partial(loop.run_in_executor, executor)
        loop.run_until_complete(
            self._fetch_large_file(run, url, filesize, data, timeout=timeout)
        )

    async def _fetch_large_file(
        self,
        run,
        url,
        file_size: int,
        data: List,
        chunk_size=10 * 1000 * 1000,
        timeout=(None, None),
    ):
        """async implementation of fetch_large_file"""

        num_chunks = file_size // chunk_size + 1
        data_array = [io.BytesIO() for i in range(num_chunks)]
        tasks = [
            run(self._fetch_range, url, i, chunk_size, data_array, timeout)
            for i in range(num_chunks)
        ]

        await asyncio.wait(tasks)

        for i in range(num_chunks):
            data_array[i].seek(0)
            data[0].write(data_array[i].read())

    async def _async_fetch_files(
        self,
        run,
        urls,
        filepaths: Union[list, str],
        etags=[],
        auto_unzip: bool = True,
        timeout=(None, None),
    ):
        """async implementation of fetch_files function"""
        tasks = []
        for idx, url in enumerate(urls):
            # get filepath
            if isinstance(filepaths, str):
                filepath = filepaths
            elif isinstance(filepaths, list) and len(filepaths) > idx:
                filepath = filepaths[idx]
            else:
                raise Exception(
                    "The 'filepaths' should be either one string or a list of strings. And the length of the list should be the same with the length of urls. "
                )

            # get etag
            if len(etags) > idx:
                etag = etags[idx]
            else:
                etag = None

            filename = None
            tasks.append(
                run(self.fetch_file, url, filepath, filename, etag, auto_unzip, timeout)
            )
        # print(tasks)
        await asyncio.wait(tasks)

    def fetch_files(
        self,
        urls,
        filepaths: Union[list, str],
        filenames=[],
        etags=[],
        auto_unzip: bool = True,
        timeout=(None, None),
    ):
        """fetch multiple files concurrently

        :param urls: the urls to download files from
        :param filepaths: location(s) to keep the files. This can be one path for all files or one path for each file.
        :param etags: old etags. if the old etag is the same with the one on server, do not download again.
        :param auto_unzip: bool flag to indicate if unzip .zip file automatically

        """

        # set up concurrent functions
        executor = concurrent.futures.ThreadPoolExecutor(max_workers=15)
        loop = asyncio.new_event_loop()
        run = functools.partial(loop.run_in_executor, executor)

        asyncio.set_event_loop(loop)

        try:
            loop.run_until_complete(
                self._async_fetch_files(
                    run,
                    urls,
                    filepaths,
                    etags=etags,
                    auto_unzip=auto_unzip,
                    timeout=timeout,
                )
            )
        except RuntimeError:
            import nest_asyncio

            nest_asyncio.apply()
            loop.run_until_complete(
                self._async_fetch_files(
                    run,
                    urls,
                    filepaths,
                    etags=etags,
                    auto_unzip=auto_unzip,
                    timeout=timeout,
                )
            )
        finally:
            loop.close()


def fetch_file(
    url: str,
    filepath: str,
    filename: Union[str, None] = None,
    etag: Union[str, None] = None,
    auto_unzip: bool = True,
    timeout=(None, None),
):
    fetcher = RequestsFetcher()
    return fetcher.fetch_file(
        url,
        filepath,
        filename=filename,
        etag=etag,
        auto_unzip=auto_unzip,
        timeout=timeout,
    )


def fetch_files(
    urls,
    filepaths: Union[list, str],
    etags=[],
    auto_unzip: bool = True,
    timeout=(None, None),
):
    fetcher = RequestsFetcher()
    return fetcher.fetch_files(
        urls, filepaths, etags=etags, auto_unzip=auto_unzip, timeout=timeout
    )


def fetch_large_file(
    url: str,
    filepath: str,
    filename: Union[str, None] = None,
    filesize: Union[int, None] = None,
    etag: Union[str, None] = None,
    auto_unzip: bool = True,
    check_etag: bool = True,
    timeout=(None, None),
):
    fetcher = RequestsFetcher()
    return fetcher.fetch_large_file(
        url,
        filepath,
        filename=filename,
        filesize=filesize,
        etag=etag,
        auto_unzip=auto_unzip,
        check_etag=check_etag,
        timeout=timeout,
    )



import argparse
import json
import logging
import os
import sys

from plate_model_manager import PlateModelManager, __version__, check_update

logger = logging.getLogger("pmm")


class ArgParser(argparse.ArgumentParser):
    def error(self, message):
        sys.stderr.write(f"error: {message}\n")
        self.print_help()
        sys.exit(1)


def _run_ls_command(args):
    if args.repository == None:
        pm_manager = PlateModelManager()
    else:
        pm_manager = PlateModelManager(args.repository)

    if args.model:
        model = pm_manager.get_model(args.model)
        if model:
            print(json.dumps(model.get_cfg(), indent=2))
        else:
            print(f"No such model {args.model}")
    else:
        for name in pm_manager.get_available_model_names():
            print(name)


def _run_download_command(args):
    logger.info(f"Downloading {args.model}")
    if args.repository == None:
        pm_manager = PlateModelManager()
    else:
        pm_manager = PlateModelManager(args.repository)

    if args.model.lower() == "all":
        pm_manager.download_all_models(args.path)
        logger.info(f"All models have been downloaded and saved in {args.path}")
    else:
        model = pm_manager.get_model(args.model)
        if model is not None:
            model.set_data_dir(args.path)
            # print(args.download_rasters)
            if args.download_rasters:
                model.download_all()
            else:
                model.download_all_layers()
            logger.info(
                f"Model({args.model}) has been downloaded and saved in {args.path}"
            )


def _run_check_update_command(args):
    check_update()


def main():
    parser = ArgParser()

    parser.add_argument("-v", "--version", action="store_true")

    subparser = parser.add_subparsers(dest="command")

    ls_cmd = subparser.add_parser(
        "ls",
        description="List all available plate model names. If given a model name, show the details of the model.",
        help="list all available plate model names. if given a model name, show the details of the model.",
    )
    ls_cmd.add_argument(
        "-r",
        "--repository",
        type=str,
        dest="repository",
        help="indicate which repository to use. you don't need this argument in most situations.",
    )
    ls_cmd.add_argument(
        "model",
        type=str,
        nargs="?",
        help="the model name. If given, show the details of the model.",
    )
    ls_cmd.set_defaults(func=_run_ls_command)

    download_cmd = subparser.add_parser(
        "download",
        description="Download a plate model or all plate models.",
        help="download a plate model or all plate models",
    )
    download_cmd.add_argument(
        "model", type=str, help="the model name. use 'all' to download all models."
    )
    download_cmd.add_argument(
        "path",
        type=str,
        nargs="?",
        default=os.getcwd(),
        help="the location to save the plate model files. use the current working directory by default.",
    )
    download_cmd.add_argument(
        "-r",
        "--repository",
        type=str,
        dest="repository",
        help="indicate which repository to use. you don't need this argument in most situations.",
    )
    download_cmd.add_argument(
        "--download-rasters",
        action="store_true",
        help="a flag to indicate if download raster files. the raster files may be large in size.",
    )
    download_cmd.set_defaults(func=_run_download_command)

    check_update_cmd = subparser.add_parser(
        "check-update",
        description="Check if new versions of plate models are available on Zenodo.",
        help="check if new versions of plate models are available on Zenodo",
    )
    check_update_cmd.set_defaults(func=_run_check_update_command)

    if len(sys.argv) == 1:
        parser.print_help(sys.stderr)
        sys.exit(0)

    args = parser.parse_args()

    if args.version:
        print(__version__)
        return

    args.func(args)


if __name__ == "__main__":
    main()



import glob
import json
import logging
import os
from hashlib import sha256
from typing import Dict

import requests

from .network_requests import fetch_file
from .utils import download, misc

DEFAULT_PRESENT_DAY_RASTERS_MANIFEST = (
    "https://repo.gplates.org/webdav/pmm/present_day_rasters.json"
)
logger = logging.getLogger("pmm")


class RasterNameNotFound(Exception):
    pass


class PresentDayRasterManager:
    """Manage the present-day rasters."""

    def __init__(self, data_dir="present-day-rasters", raster_manifest=None):
        """Constructor. Create a :class:`PresentDayRasterManager` instance.

        :param raster_manifest: The URL to a ``present_day_rasters.json`` metadata file.
                                Normally you don't need to provide this parameter unless
                                you would like to setup your own present-day raster server.

        :param data_dir: The path to a folder to save the present-day raster files.
        """
        if not raster_manifest:
            self.raster_manifest = DEFAULT_PRESENT_DAY_RASTERS_MANIFEST
        else:
            self.raster_manifest = raster_manifest
        self._rasters = None

        self.data_dir = data_dir

        # check if the model manifest file is a local file
        if os.path.isfile(self.raster_manifest):
            with open(self.raster_manifest) as f:
                self._rasters = json.load(f)
        elif self.raster_manifest.startswith(
            "http://"
        ) or self.raster_manifest.startswith("https://"):
            # try the http(s) url
            try:
                r = requests.get(self.raster_manifest)
                self._rasters = r.json()

            except requests.exceptions.ConnectionError:
                raise Exception(
                    f"Unable to fetch {self.raster_manifest}. "
                    + "No network connection or invalid URL!"
                )
        else:
            raise Exception(
                f"The model_manifest '{self.raster_manifest}' should be either a local file path or a http(s) URL."
            )

    @property
    def rasters(self) -> Dict:
        """The metadata of rasters."""
        if self._rasters is not None:
            return self._rasters
        else:
            raise Exception(
                "The self._rasters is None. This should not happen. Something Extraordinary must have happened."
            )

    @rasters.setter
    def rasters(self, var) -> None:
        self._rasters = var

    def set_data_dir(self, data_dir):
        """Set a new data folder to save the present-day rasters."""
        self.data_dir = data_dir

    def list_present_day_rasters(self):
        """Return a list of available  present-day rasters."""
        return [name for name in self.rasters]

    def _check_raster_avail(self, _name: str):
        """Check if the raster name is in raster configuration."""
        name = _name.lower()
        if not name in self.rasters:
            raise RasterNameNotFound(f"Raster {name} is not found in {self.rasters}.")
        return name

    def is_wms(self, _name: str, check_raster_avail_flag=True):
        """Return ``True`` if the raster is served by ``Web Map Service``, otherwise ``False``

        :param _name: The raster name of interest.
        :type _name: str
        :param check_raster_avail_flag: If the flag is ``True``, validate the raster name against the raster configuration.
        :type check_raster_avail_flag: bool
        """
        if check_raster_avail_flag:
            name = self._check_raster_avail(_name)
        else:
            name = _name.lower()
        if (
            isinstance(self.rasters[name], dict)
            and "service" in self.rasters[name]
            and self.rasters[name]["service"] == "WMS"
        ):
            return True
        else:
            return False

    def get_raster(
        self,
        _name: str,
        width=1800,
        height=800,
        bbox=[-180, -80, 180, 80],
        large_file_hint=True,
    ):
        """Download a raster file by name, save the raster file in ``self.data_dir`` and return the local path to the raster file.

        Call :meth:`list_present_day_rasters()` to see a list of available present-day raster names.

        :param _name: The raster name of interest.
        :type _name: str

        :return: The local path to the downloaded raster file.
        :rtype: str
        """
        name = self._check_raster_avail(_name)
        is_wms_flag = self.is_wms(name, check_raster_avail_flag=False)

        if not is_wms_flag:
            downloader = download.FileDownloader(
                self.rasters[name],
                f"{self.data_dir}/{name}/.metadata.json",
                f"{self.data_dir}/{name}/",
                large_file_hint=large_file_hint,
            )
            # only re-download when necessary
            if downloader.check_if_file_need_update():
                downloader.download_file_and_update_metadata()
            else:
                if downloader.check_if_expire_date_need_update():
                    # update the expiry date
                    downloader.update_metadata()

                logger.debug(
                    f"The local raster file {self.data_dir}/{name} is still good. Will not download again at this moment."
                )

            files = glob.glob(f"{self.data_dir}/{name}/*")
            if len(files) == 0:
                raise Exception(f"Failed to get raster {name}")
            if len(files) > 1:
                misc.print_warning(
                    f"Multiple raster files have been detected.{files}. Return the first one found {files[0]}."
                )
            return files[0]
        else:
            server_url = self.rasters[name]["server_url"]
            version = self.rasters[name]["version"]
            layers = self.rasters[name]["layers"]
            if self.rasters[name]["hillshade_layer"]:
                layers.append(self.rasters[name]["hillshade_layer"])
            styles = self.rasters[name]["styles"]
            if self.rasters[name]["hillshade_style"]:
                styles.append(self.rasters[name]["hillshade_style"])
            format = "image/geotiff"
            url = (
                f"{server_url}/wms?service=WMS&version={version}&request=GetMap&layers={','.join(layers)}"
                + f"&bbox={bbox[0]},{bbox[1]},{bbox[2]},{bbox[3]}&width={width}&height={height}&srs=EPSG:4326"
                + f"&styles={','.join(styles)}&format={format}"
            )
            filepath = (
                f"{self.data_dir}/{name}/{sha256(url.encode('utf-8')).hexdigest()}"
            )
            if not os.path.isfile(f"{filepath}/{name}.tiff"):
                fetch_file(
                    url,
                    f"{self.data_dir}/{name}/{sha256(url.encode('utf-8')).hexdigest()}",
                    filename=f"{name}.tiff",
                )
            return f"{filepath}/{name}.tiff"
